from autoslug import AutoSlugField
from django.db import models
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from gramedia.django.abstract_models import TimestampedModel


class SoftDeleteModel(models.Model):
    is_active = models.BooleanField(_('active'), default=True)
    deleted = models.DateTimeField(
        _('deleted'),
        blank=True,
        null=True,
        help_text=_('Date/time this object was deleted.')
    )

    class Meta:
        abstract = True

    @property
    def is_deleted(self):
        return True if self.deleted else False

    def delete(self):
        self.is_active = False
        self.deleted = timezone.now()
        self.save()

    def restore(self):
        self.deleted = None
        self.save()


class BaseModel(SoftDeleteModel, TimestampedModel):
    name = models.CharField(max_length=50)
    slug = AutoSlugField(populate_from='name', unique=True)

    class Meta:
        abstract = True

    def __str__(self):
        return self.name



import msgpack
from django.utils.timezone import now

from kitchenart.utils.broker import BasicPublisher, EventType


class LogPublisher(BasicPublisher):
    def construct_message(self,
                          actor: str,
                          entity: str,
                          event: EventType,
                          desc: str,
                          data_old: dict =None,
                          data_new: dict =None) -> bytes:
        """ Construct the message to be sent.
        """
        message = {
            "event_time": now().isoformat(),
            "actor": actor,
            "event": event.value,
            "entity": entity,
            "description": desc,
            "data_old": data_old,
            "data_new": data_new
        }
        return msgpack.packb(message, use_bin_type=True)

    def publish(self,
                actor: str,
                event: EventType,
                entity: str,
                desc: str,
                instance_old: any =None,
                instance_new: any =None) -> None:

        body = self.construct_message(
            actor=actor,
            entity=entity,
            event=event,
            desc=desc,
            data_old=instance_old,
            data_new=instance_new
        )

        routing_key = f'log.{entity}.{event.value}'
        self._publish(body, routing_key)



import re


class ParseIdentityHyperlink:
    def __init__(self, identity_type: str, data=None):
        """ Creates a new instance of ParseIdentityHyperlink to get
        entity identity from href (pk or slug).

        :param identity_type:  Name of the identity type. pk or slug
        :param data: list data
        """
        self.identity_type = identity_type
        self.data = data

    def get_entity_list_identity(self):
        """ Returns the ids of existing product promo that should be updated.
        """
        return [
            self.parse_identity_from_href(s['href']) for s in self.data
            if s.get('href')
        ]

    def parse_identity_from_href(self, href: str):
        """ Given an href (url string), return the entity identity.
        """
        if self.identity_type == "slug":
            regex = re.compile(r'.+?(?P<slug>[a-zA-Z0-9-_]+)/$')
        else:
            regex = re.compile(r'.+?(?P<pk>[a-zA-Z0-9-_]+)/$')

        # make sure href is always a string
        if isinstance(href, str):
            match = regex.search(href)
            if match:
                return match.groupdict()[self.identity_type]



import functools

from django.conf import settings
from django.dispatch import receiver


def custom_receiver(signal, **decorator_kwargs):
    def inner(func):
        @receiver(signal, **decorator_kwargs)
        @functools.wraps(func)
        def wrapper(sender, **kwargs):
            if settings.MUTE_SIGNALS:
                return
            return func(sender, **kwargs)
        return wrapper
    return inner



from copy import deepcopy

from rest_framework import status
from rest_framework.response import Response
from rest_framework.viewsets import ModelViewSet

from kitchenart.log.signal import user_activity_log
from kitchenart.utils.broker import EventType


class LogViewSetMixin(ModelViewSet):
    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)

        # send log signal
        user_activity_log.send(sender=self,
                               user=request.user,
                               event=EventType.created,
                               entity=self.entity,
                               instance_old=None,
                               instance_new=serializer.data)

        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)

    def update(self, request, *args, **kwargs):
        instance = self.get_object()
        instance_old = deepcopy(instance)

        result = super().update(request, *args, **kwargs)

        # data old
        serializer_old = self.get_serializer(instance_old, context={'request': request})
        data_old = serializer_old.data

        # data new
        serializer_new = self.get_serializer(self.get_object(), context={'request': request})
        data_new = serializer_new.data

        # send log signal
        user_activity_log.send(sender=self,
                               user=request.user,
                               event=EventType.changed,
                               entity=self.entity,
                               instance_old=data_old,
                               instance_new=data_new)

        return result

    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()
        self.perform_destroy(instance)

        serializer = self.get_serializer(instance, context={'request': request})

        # send log signal
        user_activity_log.send(sender=self,
                               user=request.user,
                               event=EventType.deleted,
                               entity=self.entity,
                               instance_old=serializer.data,
                               instance_new=None)

        return Response(status=status.HTTP_204_NO_CONTENT)



from enum import Enum
from typing import Type

import msgpack
from django.conf import settings
from django.contrib.sites.models import Site
from django.utils.timezone import now
from kombu import Exchange, Connection, Producer
from rest_framework.serializers import BaseSerializer


class EventType(Enum):
    created = 'created'
    changed = 'changed'
    deleted = 'deleted'
    restored = 'restored'


class BasicPublisher:
    """ Object used for publishing.
    """

    def __init__(self, exchange_name: str, site: Site):
        """ Creates a new instance of BasicPublisher.
        If connection or channel are not defined, they will be initialized
        automatically from the global publish connection and channel objects.

        :param exchange_name:  Name of the exchange this publisher should publish to.
        """
        self.exchange_name = exchange_name
        self.site = site

    def exchange(self, exchange_type='topic'):
        """ Creates an exchange
        """
        exchange = Exchange(self.exchange_name, type=exchange_type, durable=True)
        return exchange

    def get_identity(self, data: dict) -> str:
        return data['href']

    def serializer_context(self) -> dict:
        """ Simulates a request context sot hat our serializers can operate properly.

        :param site_id: Optional parameter (but should generally
        :return: A single-key dictionary containing 'request' with a simulated request object.
        """
        # this is a terrible hack to make the requests still work from the CLI.
        from rest_framework.request import Request, HttpRequest

        class SimulatedHTTPRequest(HttpRequest):
            def _get_scheme(self):
                return "https"

        req = SimulatedHTTPRequest()

        req.META = {
            'SERVER_NAME': self.site.domain,
            'SERVER_PORT': 443,
            'SCRIPT_NAME': settings.FORCE_SCRIPT_NAME
        }
        return {'request': Request(req)}

    def _publish(self, body: any, routing_key: str) -> None:
        exchange = self.exchange()
        with Connection(settings.BROKER_URL) as conn:
            with conn.channel() as channel:
                producer = Producer(channel)
                producer.publish(
                    body,
                    exchange=exchange,
                    declare=[exchange],
                    content_type='application/x-msgpack',
                    routing_key=routing_key,
                    retry=True,
                    retry_policy={
                        'interval_start': 0,
                        'interval_step': 2,
                        'interval_max': 30,
                        'max_retries': 30
                    }
                )


class EntityPublisher(BasicPublisher):
    def construct_message(self, data: dict, entity: str, event: EventType, identity=None) -> bytes:
        """ Construct the message to be sent.
        """
        message = {
            "event_time": now().isoformat(),
            "identity": identity or self.get_identity(data),
            "event": event.value,
            "entity": entity,
            "data": data
        }
        return msgpack.packb(message, use_bin_type=True)

    def publish(self,
                message: any,
                message_serializer: Type[BaseSerializer],
                event: EventType,
                entity: str,
                message_identity: any = None) -> None:
        """ Publish the message to RabbitMQ.
        """
        data = message_serializer(message, context=self.serializer_context()).data

        body = self.construct_message(
            data=data,
            entity=entity,
            event=event,
            identity=message_identity
        )

        routing_key = f'{entity}.{event.value}'
        self._publish(body, routing_key)



import codecs
import os
import re

from setuptools import setup, find_packages


def read(*parts):
    file_path = os.path.join(os.path.dirname(__file__), *parts)
    return codecs.open(file_path, encoding='utf-8').read()


def find_version(*parts):
    version_file = read(*parts)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return str(version_match.group(1))
    raise RuntimeError("Unable to find version string.")


with open('README.md') as readme_file:
    README = readme_file.read()

with open('HISTORY.md') as history_file:
    HISTORY = history_file.read()

setup(
    name='ka-drf-utils',
    version=find_version('kitchenart', '__init__.py'),
    license='MIT',
    author='MSJ Development Team',
    author_email='engineering@muliasuksesjaya.com',
    description='KitchenArt utility package',
    long_description_content_type="text/markdown",
    long_description=README + '\n\n' + HISTORY,
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'gdn-python-common[drf]>=0.8.9',
    ],
    extras_require={
        'broker': [
            'kombu>=5.0',
            'msgpack==1.0.0',
        ]
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



from django.conf import settings
from django.contrib.sites.models import Site

from kitchenart import _logger
from kitchenart.log.publisher import LogPublisher
from kitchenart.utils.broker import EventType


def publish_activity_log(sender, user, event, entity, instance_old, instance_new, **kwargs):
    actor = f'{user.get_full_name()} <{user.email}>'
    description = None
    entity_name = entity.replace('_', ' ')
    if event == EventType.created:
        description = f'Create new {entity_name}: {instance_new}'
    elif event == EventType.changed:
        description = f'Update {entity_name}: {instance_new}'
    elif event == EventType.deleted:
        description = f'Delete {entity_name}: {instance_old}'

    try:
        exchange_name = settings.LOG_EXCHANGE
        publisher = LogPublisher(exchange_name, site=Site.objects.get_current())
        publisher.publish(
            actor=actor,
            event=event,
            entity=entity,
            desc=description,
            instance_old=instance_old,
            instance_new=instance_new
        )
    except Exception as exc:
        _logger.exception(f"AMQP error log publish {entity}: {exc}")



from http import HTTPStatus

from django.conf import settings
from gramedia.django.drf import convert_env_boolean
from rest_framework import serializers
from rest_framework.exceptions import APIException
from rest_framework.response import Response
from rest_framework.views import exception_handler

from kitchenart import _logger


class TimestampedSerializerMixin:
    """ Automatically attaches the modified_by to a model.

    .. warning::
        This is only for use with Timestemped models, with an attribute 'modified_by'
    """
    def modified_user(self):
        user = self.context['request'].user
        return f'{user.get_full_name()} <{user.email}>'

    def create(self, validated_data):
        validated_data['modified_by'] = self.modified_user()
        return super().create(validated_data)

    def update(self, instance, validated_data):
        validated_data['modified_by'] = self.modified_user()
        return super().update(instance, validated_data)


class SoftDeleteViewSetMixin:
    def get_queryset(self):
        if convert_env_boolean(self.request.query_params.get('include_deleted', '')):
            return self.queryset
        else:
            return self.queryset.filter(deleted__isnull=True)


def get_entity_href_serializer(model_class, meta_extra_kwargs=None, *init_args, **init_kwargs):
    class EntityHrefSerializer(serializers.HyperlinkedModelSerializer):
        name = serializers.CharField(required=False)

        class Meta:
            model = model_class
            fields = ('href', 'name',)
            extra_kwargs = meta_extra_kwargs or {'href': {'lookup_field': 'slug', }, }
    return EntityHrefSerializer(*init_args, **init_kwargs)


def custom_exception_handler(exc, context):
    # Call REST framework's default exception handler first,
    # to get the standard error response.
    response = exception_handler(exc, context)

    if response is not None:
        response.data = {}
        if hasattr(exc, 'detail') and hasattr(exc.detail, 'items'):
            response.data['detail'] = exc.detail

        if isinstance(exc, APIException):
            response.data['message'] = exc.detail if isinstance(exc.detail, str) else exc.default_detail
        else:
            response.data['message'] = str(exc)
    else:
        _logger.error(exc)
        if not settings.DEBUG:
            response = Response({
                'message': HTTPStatus.INTERNAL_SERVER_ERROR.phrase,
                'detail': []
            }, status=HTTPStatus.INTERNAL_SERVER_ERROR)

    return response



from gramedia.common.http import LinkHeaderField, LinkHeaderRel
from gramedia.django.drf import LinkHeaderPagination as GramediaLinkHeaderPagination
from rest_framework.response import Response


class LinkHeaderPagination(GramediaLinkHeaderPagination):
    def get_paginated_response(self, data):

        next_url = self.get_next_link()
        previous_url = self.get_previous_link()

        link_parts = []

        if next_url is not None:
            link_parts.extend([
                LinkHeaderField(
                    url=next_url,
                    rel=LinkHeaderRel.next,
                    title=str(self.page.number + 1)),
                LinkHeaderField(
                    url=self.get_last_url(),
                    rel=LinkHeaderRel.last,
                    title=self.page.paginator.num_pages),
            ])

        # if we have a previous page, then we know we have a first page, too.
        if previous_url is not None:
            link_parts.extend([
                LinkHeaderField(
                    url=previous_url,
                    rel=LinkHeaderRel.prev,
                    title=str(self.page.number - 1)),
                LinkHeaderField(
                    url=self.get_first_url(),
                    rel=LinkHeaderRel.first,
                    title=str(1)),
            ])

        headers = {
            'X-Total-Results': self.page.paginator.count,
            'X-Page-Size': self.get_page_size(self.request),
            'X-Page': self.page.number
        }

        if link_parts:
            headers['Link'] = ", ".join([str(link) for link in link_parts])

        return Response(data, headers=headers)
